#include "phys.h"

particle :: particle()
{
	nextParticle = NULL;
	position[X] = 0; position[Y] = 0; position[Z] = 0;
	velocity[X] = 0; velocity[Y] = 0; velocity[Z] = 0;
	forceAccumulator[X] = 0;
	forceAccumulator[Y] = 0; forceAccumulator[Z] = 0;
	mass = 0;
	clamped = FALSE;
}

particle :: ~particle()
{
	if (nextParticle) delete nextParticle;
	cout << "Particle deleted" << endl;
}


void particle :: AttachParticle(particle *aParticle)
{
	nextParticle = aParticle;
}

void particle :: AssignMass(float m)
{
	mass = m;
}

float particle :: GetMass()
{
	return mass;
}

void particle :: AssignPosition(float x, float y, float z)
{
	position[X] = x; position[Y] = y; position[Z] = z;
}

void particle :: AssignPosition(float pos[3])
{
	position[X] = pos[X]; position[Y] = pos[Y]; position[Z] = pos[Z];
}

void particle :: GetPosition(float &x, float &y, float &z)
{
	x = position[X]; y = position[Y]; z = position[Z];
}

void particle :: GetPosition(float pos[3])
{
	pos[X] = position[X]; pos[Y] = position[Y]; pos[Z] = position[Z];
}

float * particle :: GetPosition()
{
	return position;
}


void particle :: AssignVelocity(float x, float y, float z)
{
	velocity[X] = x; velocity[Y] = y; velocity[Z] = z;
}

void particle :: AssignVelocity(float vel[3])
{
	velocity[X] = vel[X]; velocity[Y] = vel[Y]; velocity[Z] = vel[Z];
}

void particle :: GetVelocity(float &x, float &y, float &z)
{
	x = velocity[X]; y = velocity[Y]; z = velocity[Z];
}

void particle :: GetVelocity(float vel[3])
{
	vel[X] = velocity[X]; vel[Y] = velocity[Y]; vel[Z] = velocity[Z];
}

float * particle :: GetVelocity() {return velocity;}



void particle :: AssignForce(float x, float y, float z)
{
	forceAccumulator[X] = x; forceAccumulator[Y] = y; forceAccumulator[Z] = z;
}

void particle :: AssignForce(float force[3])
{
	forceAccumulator[X] = force[X]; forceAccumulator[Y] = force[Y]; forceAccumulator[Z] = force[Z];
}

void particle :: AddForce(float x, float y, float z)
{
	forceAccumulator[X] += x; forceAccumulator[Y] += y;
	forceAccumulator[Z] += z;
}

void particle :: AddForce(float force[3])
{
	forceAccumulator[X] += force[X];
	forceAccumulator[Y] += force[Y]; forceAccumulator[Z] += force[Z];
}


void particle :: GetForce(float &x, float &y, float &z)
{
	x = forceAccumulator[X]; y = forceAccumulator[Y];
	z = forceAccumulator[Z];
}

void particle :: GetForce(float force[3])
{
	force[X] = forceAccumulator[X];
	force[Y] = forceAccumulator[Y]; force[Z] = forceAccumulator[Z];
}

int particle :: IsClamped()
{
	if (clamped) return TRUE;
	else return FALSE;
}

particle * particle :: NextParticle()
{
	return nextParticle;
}


/*****************************************************************/


particleSystem :: particleSystem()
{
	listOfParticles = NULL;
	numberOfParticles = 0;
	lastParticle = NULL;
	time = 0;
}

particleSystem :: ~particleSystem()
{
	if (listOfParticles) delete listOfParticles;
}

void particleSystem :: AddParticle(particle *aParticle)
{
	if (listOfParticles == NULL) {
		listOfParticles = aParticle;
		lastParticle = aParticle;
	} else {
		lastParticle->AttachParticle(aParticle);
		lastParticle = aParticle;
	}
	numberOfParticles++;
}

void particle :: ClearForceDeep()
{
	forceAccumulator[X] = 0;
	forceAccumulator[Y] = 0;
	forceAccumulator[Z] = 0;
	if (nextParticle) nextParticle->ClearForceDeep();
}

void particleSystem :: ClearForces()
{
	listOfParticles->ClearForceDeep();
}


particle * particleSystem :: GetNthParticle(int n)
{

	if ((n < 0) || (n >= numberOfParticles)) {
		cerr << "GetNthParticle: Particle " << n << " does not exist." << endl;
		exit(1);
	}

	particle *eachParticle = listOfParticles;
	int count = 0;

	while(count != n) {
		eachParticle = eachParticle->NextParticle();
		count++;
	}
	return eachParticle;
}


particle *particleSystem :: FirstParticle()
{
	return listOfParticles;
}
void ComputeForceDueToGravity(particleSystem *ps)
{
	particle *eachParticle;

	eachParticle = ps->FirstParticle();

#ifdef DEBUG
	cout << "Computing force due to gravity" << endl;
#endif
	while(eachParticle) {

		/* Gravity is F = mg where g is 9.8 m.s.s but it only
		 * acts in the -Y direction.
		 * We add this force to the force accumulators of each
		 * particle. By adding we can compute any of the forces
		 * in any order we choose.
		 */

		eachParticle->AddForce(0.0,
				       eachParticle->GetMass() * -9.8,
				       0.0);

		eachParticle = eachParticle->NextParticle();
	}
}


float Dot(float x1, float y1, float z1, float x2, float y2, float z2)
{
#ifdef DEBUG
	cout << "DOT: " << x1 * x2 + y1 * y2 + z1 * z2 << endl;
#endif
	return x1 * x2 + y1 * y2 + z1 * z2;
}

float Dot(float v1[3], float v2[3])
{
	return Dot(v1[X],v1[Y],v1[Z],v2[X],v2[Y],v2[Z]);
}

int Collided(particle *aParticle, bouncePlane *aPlane)
{
	/* (X-P).N where X is point. P is point on plane, N is normal of plane */


	float px,py,pz, cx,cy,cz, nx, ny, nz;

#ifdef DEBUG
	cout << "Checking for collisions" << endl;
#endif
	aParticle->GetPosition(px,py,pz);
	aPlane->GetNormal(nx,ny,nz);
	aPlane->GetCorner1(cx,cy,cz);

	if  (Dot(px-cx,py-cy,pz-cy,nx,ny,nz) < 0) {
		return TRUE;
	} else return FALSE;

}

void Bounce(particle *aParticle, bouncePlane *aPlane)
{
	float vnx, vny, vnz;
	float vtx, vty, vtz;
	float dot, vx, vy, vz;

	/* Compute normal and tangential components of velocity
	 * Vn = (V . N)N  and Vt = V - Vn  p.407 of Karl Sims SIGGRAPH 90
	 */

	dot= Dot(aParticle->GetVelocity(), aPlane->GetNormal());

	aPlane->GetNormal(vnx,vny,vnz);
	vnx *= dot;
	vny *= dot;
	vnz *= dot;

	aParticle->GetVelocity(vtx,vty,vtz);
	vtx -= vnx;
	vty -= vny;
	vtz -= vnz;

	/* New velocity = (1 - u) Vt - e Vn
	 * where u = friction and e = resilience of bounce
	 */

	float friction = aPlane->GetFriction();
	float resilience = aPlane->GetResilience();
	aParticle->AssignVelocity((1.0 - friction) * vtx -
				  resilience * vnx,
		       (1.0 - friction) * vty - resilience * vny,
		       (1.0 - friction) * vtz - resilience * vnz);
}


void ComputeBounce(particleSystem *ps, bounceSurfaces *bs)
{

	particle *eachParticle = ps->FirstParticle();
	bouncePlane *eachPlane;

#ifdef DEBUG
	cout << "Compute bounce" << endl;
#endif

	while (eachParticle) {
		eachPlane = bs->FirstBouncePlane();
		while (eachPlane) {
			if (Collided(eachParticle, eachPlane)) {
#ifdef DEBUG
				cout << ""<<endl;
#endif

				Bounce(eachParticle, eachPlane);
			}
			eachPlane = eachPlane->NextBouncePlane();
		}
		eachParticle = eachParticle->NextParticle();

	}
}

float Magnitude(float x1,float x2, float y1, float y2, float z1, float z2)
{
	float x = x1-x2, y = y1-y2, z = z1-z2;
	return sqrt(x*x + y*y + z*z);
}

float Magnitude(float p1[3], float p2[3])
{
	return Magnitude(p1[X],p2[X],p1[Y],p2[Y],p1[Z],p2[Z]);
}



void ComputeSpringForces(particleSystem *ps, connections *cons)
{

	float restingLength;

	float springConstant, springDamping;

	spring *eachSpring = cons->FirstSpring();
	particle *partA, *partB;

	float posMag;
	float pax,pay,paz, pbx, pby, pbz, vax, vay, vaz, vbx, vby, vbz;


	float fax, fay, faz;

	float fa;
	while(eachSpring) {
		restingLength = eachSpring->GetRestingDistance();

		partA = eachSpring->GetHead();
		partB = eachSpring->GetTail();

		partA->GetPosition(pax,pay,paz);
		partB->GetPosition(pbx,pby,pbz);
		partA->GetVelocity(vax,vay,vaz);
		partB->GetVelocity(vbx,vby,vbz);

		posMag = Magnitude(pax,pbx,pay,pby,paz,pbz);

		springConstant = eachSpring->GetKs();
		springDamping = eachSpring->GetKd();

		fa = -(springConstant * (posMag - restingLength) +
		       springDamping * 
		       ((Dot(vax-vbx,vay-vby,vaz-vbz,pax-pbx,
			     pay-pby,paz-pbz) / posMag)));

		fax = fa/posMag * (pax-pbx);
		fay = fa/posMag * (pay-pby);
		faz = fa/posMag * (paz-pbz);

		partA->AddForce(fax,fay,faz);
		partB->AddForce(-fax,-fay,-faz);

		eachSpring = eachSpring->NextSpring();
	}

}

void ComputeForces(particleSystem *ps, bounceSurfaces *bs, connections *cons)
{
#ifdef DEBUG
	cout << "Computing forces" << endl;
#endif
	ComputeForceDueToGravity(ps);
	ComputeSpringForces(ps,cons);
	ComputeBounce(ps,bs);
}


void particleSystem :: EulerStep(float deltaTime, bounceSurfaces *bs, connections *cons)
{

	/* First clear all force accumulators to zero. */
	ClearForces();

	/* Compute the forces on the current system */
	ComputeForces(this,bs,cons);

	particle *eachParticle = listOfParticles;

	float v1x, v1y, v1z;
	float v2x, v2y, v2z;
	float fx, fy, fz, ax, ay, az;
	float px,py,pz;

	/* go through and update each particle using the new forces */
	while(eachParticle) {

		/* Get current velocity and force of a particle*/
		eachParticle->GetVelocity(v1x, v1y, v1z);
		eachParticle->GetForce(fx, fy, fz);

		/* Compute acceleration of the particle.
		 * F=MA so A=F/M
		 */
		ax = fx/eachParticle->GetMass();
		ay = fy/eachParticle->GetMass();
		az = fz/eachParticle->GetMass();

		/* Scale these values by delta Time to produce
		 * deltaPosition and deltaVelocity
		 */
		v1x *= deltaTime;
		v1y *= deltaTime;
		v1z *= deltaTime;
		
		ax *= deltaTime;
		ay *= deltaTime;
		az *= deltaTime;

		/* Get current state of the particle:
		 * position and velocity.
		 */
		eachParticle->GetPosition(px,py,pz);
		eachParticle->GetVelocity(v2x,v2y,v2z);

		/* Update position and velocity using previous
		 * scaled velocity and acceleration (i.e. deltaPosition
		 * and deltaVelocity respectively).
		 */
		eachParticle->AssignPosition(px+v1x, py+v1y, pz+v1z);
		eachParticle->AssignVelocity(v2x+ax, v2y+ay, v2z+az);

		eachParticle = eachParticle->NextParticle();
		time += deltaTime;
	}


}

particle * spring :: GetHead() {return head;}
particle * spring :: GetTail() {return tail;}

spring * connections :: FirstSpring() {return listOfSprings;}

void spring :: AssignKs(float k)
{	ks = k;
}

void spring :: AssignKd(float k)
{	kd = k;
}
void spring :: AssignHead(particle *p)
{
	head = p;
}

float spring :: GetKs() {return ks;}
float spring :: GetKd() {return kd;}

void spring :: AssignTail(particle *p)
{
	tail = p;
}

void spring :: AssignRestingDistance(float r)
{	restingDistance = r;
}

float spring :: GetRestingDistance() {return restingDistance;}

void connections :: Connect(particle *one, particle *two, float springConstant, float springDamping)
{
	spring *newSpring;

	newSpring = new spring;

	newSpring->AssignKs(springConstant);
	newSpring->AssignKd(springDamping);
	newSpring->AssignHead(one);
	newSpring->AssignTail(two);
	newSpring->AssignRestingDistance(Magnitude(one->GetPosition(), two->GetPosition()));

	if (listOfSprings == NULL) {
		listOfSprings = newSpring;
		lastSpring = newSpring;
	} else {
		lastSpring->AddSpring(newSpring);
		lastSpring = newSpring;
	}
	numberOfConnections++;
}

void connections :: Connect(particleSystem *ps, int one, int two, float springConstant, float springDamping)
{
	particle *first, *second;

	first = ps->GetNthParticle(one);
	second = ps->GetNthParticle(two);
	Connect(first,second,springConstant,springDamping);
}

spring * spring :: NextSpring()
{
	return nextSpring;
}

void spring :: AddSpring(spring *newSpring)
{
	nextSpring = newSpring;
}

spring :: spring()
{
	head = NULL;
	tail = NULL;
	nextSpring = NULL;
}

connections :: connections()
{
	listOfSprings = NULL;
	lastSpring = NULL;
	numberOfConnections = 0;
}

spring * connections :: GetNthConnection(int n)
{

	if ((n < 0) || (n >= numberOfConnections)) {
		cerr << "GetNthConnection: Spring connection " << n << " does not exist." << endl;
		exit(1);
	}

	spring *eachSpring = listOfSprings;
	int count = 0;

	while(count != n) {
		eachSpring = eachSpring->NextSpring();
		count++;
	}
	return eachSpring;
}

connections :: ~connections()
{
	cout << "Have not implemented destructor for connections yet" << endl;
}

bouncePlane :: bouncePlane()
{
	corner1[X] = 0;corner1[Y] = 0;corner1[Z] = 0;
	corner2[X] = 0;corner2[Y] = 0;corner2[Z] = 0;
	corner3[X] = 0;corner3[Y] = 0;corner3[Z] = 0;
	nextBouncePlane = NULL;
	normal[X] = 0; normal[Y] = 0; normal[Z] = 0;
}

bouncePlane :: bouncePlane(float x1, float y1, float z1,
			     float x2, float y2, float z2,
			     float x3, float y3, float z3, float fric, float res)
{
	corner1[X] = x1;corner1[Y] = y1;corner1[Z] = z1;
	corner2[X] = x2;corner2[Y] = y2;corner2[Z] = z2;
	corner3[X] = x3;corner3[Y] = y3;corner3[Z] = z3;	
	u = fric; e = res;
	nextBouncePlane = NULL;
	ComputeNormal();
}

bouncePlane :: bouncePlane(float p1[3], float p2[3], float p3[3], float fric, float res)
{
	corner1[X] = p1[X];corner1[Y] = p1[Y];corner1[Z] = p1[Z];
	corner2[X] = p2[X];corner2[Y] = p2[Y];corner2[Z] = p2[Z];
	corner3[X] = p3[X];corner3[Y] = p3[Y];corner3[Z] = p3[Z];	
	u = fric; e = res;
	nextBouncePlane = NULL;
	ComputeNormal();
}

void bouncePlane :: ComputeNormal()
{
	float v1,v2,v3,w1,w2,w3,nx,ny,nz,size;

	v1 = corner2[X] - corner1[X];
	v2 = corner2[Y] - corner1[Y];
	v3 = corner2[Z] - corner1[Z];

	w1 = corner3[X] - corner2[X];
	w2 = corner3[Y] - corner2[Y];
	w3 = corner3[Z] - corner2[Z];

	nx = v2 * w3 - v3 * w2;
	ny = v3 * w1 - v1 * w3;
	nz = v1 * w2 - v2 * w1;
	size = sqrt(nx * nx + ny * ny + nz * nz);

	normal[X] = nx/size; normal[Y] = ny/size; normal[Z] = nz/size;
	cout << "normal " << normal[X] << " " << normal[Y] << " " << normal[Z] << endl;
}

void bouncePlane :: GetNormal(float &x, float &y, float &z)
{
	x = normal[X]; y = normal[Y]; z = normal[Z];
}

void bouncePlane :: GetNormal(float norm[3])
{
	norm[X] = normal[X];
	norm[Y] = normal[Y];
	norm[Z] = normal[Z];
}

float * bouncePlane :: GetNormal() {return normal;}

bouncePlane * bouncePlane :: NextBouncePlane()
{
	return nextBouncePlane;
}


bouncePlane * bounceSurfaces::FirstBouncePlane()
{
	return listOfPlanes;
}

void bouncePlane :: AssignFriction(float f) {u = f;}
void bouncePlane :: AssignResilience(float r) {e = r;}
float bouncePlane :: GetFriction() {return u;}
float bouncePlane :: GetResilience() {return e;}

void bounceSurfaces :: AddSurface(float x1, float y1, float z1,
			     float x2, float y2, float z2,
			     float x3, float y3, float z3, float fric, float res)
{
	bouncePlane *newPlane = new bouncePlane(x1,y1,z1,x2,y2,z2,x3,y3,z3,fric,res);
	
	if (listOfPlanes == NULL) {
		listOfPlanes = newPlane;
		lastPlane = newPlane;
	} else {
		lastPlane->AddPlane(newPlane);
		lastPlane = newPlane;
	}
	numberOfPlanes++;
}

void bouncePlane :: AddPlane(bouncePlane *newPlane)
{
	nextBouncePlane = newPlane;
}


void bouncePlane :: GetCorner1(float &x, float &y, float &z)
{
	x = corner1[X]; y = corner1[Y], z= corner1[Z];
}

bounceSurfaces :: bounceSurfaces()
{
	listOfPlanes = NULL;
	lastPlane = NULL;
	numberOfPlanes = 0;
}
bounceSurfaces :: ~bounceSurfaces()
{	cout << "need destructor for bouncesurfaces" << endl;
}



/* Notes for implementing midpoint method
 * f1  = delta_t  f(x)
 * delta_x = delta_t f(x + f1/2)
 */
