#include <stdlib.h>
#include <iostream.h>
#include <math.h>

const int X=0;
const int Y=1;
const int Z=2;
const int TRUE=1;
const int FALSE=0;




class bouncePlane {
      protected:
	float corner1[3], corner2[3], corner3[3];
	float normal[3];
	bouncePlane *nextBouncePlane;
	float u, e;	/* Friction and resilience */
      public:

	bouncePlane();
	bouncePlane(float,float,float,float,float,float,float,float,float,float,float);
	bouncePlane(float *, float*, float*, float, float);
	~bouncePlane();

	void AssignFriction(float);
	void AssignResilience(float);
	float GetFriction();
	float GetResilience();

	void ComputeNormal();
	void GetNormal(float &,float &,float &);
	void GetNormal(float *);
	float *GetNormal();
	bouncePlane *NextBouncePlane();
	void GetCorner1(float &, float &, float &);
	void AddPlane(bouncePlane *);
};

class bounceSurfaces {
      protected:
	bouncePlane *listOfPlanes;
	bouncePlane *lastPlane;
	int numberOfPlanes;

      public:
	bounceSurfaces();
	~bounceSurfaces();
	bouncePlane * FirstBouncePlane();
	void AddSurface(float,float,float,float,float,float,float,float,float,float,float);

};



class particle {
      protected:
	float mass;
	float position[3];
	float velocity[3];
	float forceAccumulator[3];
	particle *nextParticle;
	int clamped;

      public:

	particle();
	~particle();

	void AttachParticle(particle *);

	int IsClamped();
	float GetMass();
	void AssignMass(float);
	void AssignPosition(float,float,float);
	void AssignPosition(float pos[3]);
	void GetPosition(float &x, float &y, float &z);
	void GetPosition(float pos[3]);
	float * GetPosition();
	void AssignVelocity(float,float,float);
	void AssignVelocity(float vel[3]);
	void GetVelocity(float &x, float &y, float &z);
	void GetVelocity(float vel[3]);
	float * GetVelocity();
	void AssignForce(float,float,float);
	void AssignForce(float force[3]);
	void AddForce(float,float,float);
	void AddForce(float force[3]);
	void GetForce(float &x, float &y, float &z);
	void GetForce(float force[3]);
	particle *NextParticle();
	void ClearForceDeep();

};



class spring {
      protected:
	particle *head;
	particle *tail;
	float ks,kd;	/* Spring constant and damping */
	float restingDistance;
	spring *nextSpring;

      public:
	spring();
	~spring();
	void AssignHead(particle *);
	void AssignTail(particle *);
	void AssignKs(float);
	void AssignKd(float);
	void AssignRestingDistance(float);
	float GetRestingDistance();
	float GetKs();
	float GetKd();
	void AddSpring(spring *);
	particle *GetHead();
	particle *GetTail();
	spring * NextSpring();
};


class connections;

class particleSystem {

      protected:
	particle *lastParticle;
	particle *listOfParticles;
	int numberOfParticles;
	float time;	/* Simulation Clock */

      public:

	particleSystem();
	~particleSystem();
	void AddParticle(particle *);
	void ClearForces();
	void EulerStep(float, bounceSurfaces *, connections *);
	particle * GetNthParticle(int);
	particle *FirstParticle();
};


class connections {

      protected:
	spring *listOfSprings;
	spring *lastSpring;
	int numberOfConnections;

      public:
	connections();
	~connections();
	void Connect(particle *, particle *, float, float);
	void Connect(particleSystem *, int, int, float, float);
	spring *GetNthConnection(int);
	spring *FirstSpring();
};
